#!/usr/bin/perl

use warnings;
use Getopt::Long;

my $snpsfile = '';	#file paht to snps.txt
my $gfffile = '';	#file path to gff file
my $opt_help = 0;
my $prefix_dir = 'snpmatch';	#prefixed name of putput directory
my $position = 1;		#position of column where snps positions are written in snps.txt
my $chr = '';		#name of target chromosome in gff file
my @feature;		#list of target features in gff file
my $source = "";	#source of intron donor/accepter
my $donor = 4;		#length of intron donor
my $accepter = 4;	#length of intron accepter


my $log;
GetOptions('target=s' => \$snpsfile,'gfffile=s' => \$gfffile,
'help' => \$opt_help, 'chr=s' => \$chr,'feature=s' => \@feature,'position=i' => \$position, 
'intron=s' => \$source,'donor=i' => \$donor, 'accepter=i' => \$accepter);

#Check options and input files
if(!$snpsfile or !$gfffile or $opt_help or !$chr){
	die "USAGE:snpsgff.pl -t /path/to/snps.txt -g path/to/gfffile -c chrN
Options:
-t <string>	path to snps.txt file from CASAVA.(essntial)
-g <string>	path to gfffiles (essentail)
-c <integer>	name of target chromosome in 1st column of gff file
-p <integer>	position of column of SNPs position default:1
-f <string>	list of target features in gff file.
-i <string> source feature of intron donor/accepter
-d <integer>length of intron donor. default:4
-a <integer>length of intron accepter default:4
-h 	help. show this message\n"
}

if ($position < 0){
	die "ERROR:column of SNPs positions (option:-p) is invalid value\n";
}else{
printf STDERR "column of SNPs positions:%d\n",$position;
}
if(!@feature && !$source){
	@feature =(five_prime_UTR,CDS,three_prime_UTR);
}

print STDERR "Target features:@feature\n";
$log .= "Target features:@feature\n";
print STDERR "Target chromosome:$chr\n";
$log .= "Target chromosome:$chr\n";

#READING gff file from TAIR
open(IN, "$gfffile") or die "Can't open file:$gfffile\n";
while(<IN>){
	$gff_raw .= $_;
}
close(IN);
$gff_raw =~ s/\x0D\x0A|\x0D|\x0A/\n/g;
@gff_data = split(/\n/,$gff_raw);
undef $gff_raw;
printf STDERR "gff row num:%s\n",$#gff_data;
$log .= sprintf "gff row num:%s\n",$#gff_data;
#EXTRACT feature data from raw GFF data

foreach $feat (@feature){
	my $pattern = "^$chr\t\[^\t]+\t$feat";
	@{$feat.'data'} = grep(/$pattern/,@gff_data);
	printf STDERR "gff %s num:%d\n",$feat,$#{$feat.'data'}+1;
	$log .= sprintf "gff %s num:%d\n",$feat,$#{$feat.'data'}+1;
}


#MAKING intron donor/accepter data
if($source){
	my $pattern = "^$chr\t[^\t]+\t$source";
	my @i_source;
	printf STDERR "source of intron data:%s\n",$source;
	$log .= sprintf "source of intron data:%s\n",$source;

	@i_source = grep(/$pattern/,@gff_data);
	foreach (@i_source){
		my $name;
		$name = (split(/\t/,$_))[8];
		push(@mRNA_list,$name);
	}
	my %countgrep;
	@mRNA_list = grep {!$countgrep{$_}++} @mRNA_list;
	printf STDERR "num of mRNA list:%d\n",$#mRNA_list;
	$log .= sprintf "num of mRNA list:%d\n",$#mRNA_list;	


	($sec,$min,$hour) = (localtime(time))[0,1,2];
	printf STDERR "loop start at %02d:%02d:%02d\n",$hour,$min,$sec;
	printf STDERR "Making intron donor/accepter data. Wait several minutes\n";

	$log .= sprintf "loop start at %02d:%02d:%02d\n",$hour,$min,$sec;
	$log .= sprintf "Making intron donor/accepter data. Wait several minutes\n";
	
	foreach $mRNA (@mRNA_list){
		@da_source = grep(/$mRNA/,@i_source);

		if($#da_source > 0){
			my $start;
			my %end;
			my %start;
			my %strand;
			my %name;
			
			foreach(@da_source){
				$start=(split(/\t/,$_))[3];
				($end{$start},$strand{$start},$name{$start})=(split(/\t/,$_))[4,6,8];
			}
			
			my @sort_key = sort {$a <=> $b} keys %end;
			
			for($i = 0; $i <= $#sort_key; $i++){
				my $for = "donor";
				my $rev = "accepter";
				my $l_for = $donor;
				my $l_rev = $accepter;
				if($strand{$sort_key[$i]} eq "-"){
					$for = "accepter";
					$rev = "donor";
					$l_for = $accepter;
					$l_rev = $donor;
				}elsif($strand{$sort_key[$i]} ne "+"){
					print STDERR "Warnning:$sort_key[$i]-$end{$sort_key[$i]} does not have direction information\n";
				}
				if($i == 0){
					${$for} = sprintf "%s\tsnpsmatch\tintron_%s\t%d\t%d\t.\t%s\t.\t%s",$chr,$for,$end{$sort_key[$i]}+1,$end{$sort_key[$i]}+$l_for,$strand{$sort_key[$i]},$name{$sort_key[$i]};
					push(@intron_donor_accepterdata,${$for});
				}elsif($i == $#sort_key){
					${$rev} = sprintf "%s\tsnpsmatch\tintron_%s\t%d\t%d\t.\t%s\t.\t%s",$chr,$rev,$sort_key[$i]-$l_rev,$sort_key[$i]-1,$strand{$sort_key[$i]},$name{$sort_key[$i]};
					push(@intron_donor_accepterdata,${$rev});
				}else{
					${$rev} = sprintf "%s\tsnpsmatch\tintron_%s\t%d\t%d\t.\t%s\t.\t%s",$chr,$rev,$sort_key[$i]-$l_rev,$sort_key[$i]-1,$strand{$sort_key[$i]},$name{$sort_key[$i]};
					${$for} = sprintf "%s\tsnpsmatch\tintron_%s\t%d\t%d\t.\t%s\t.\t%s",$chr,$for,$end{$sort_key[$i]}+1,$end{$sort_key[$i]}+$l_for,$strand{$sort_key[$i]},$name{$sort_key[$i]};
					push(@intron_donor_accepterdata,${$rev});
					push(@intron_donor_accepterdata,${$for});
				}
			}
			
		}
	}
	($sec,$min,$hour) = (localtime(time))[0,1,2];
	printf STDERR "loop end at %02d:%02d:%02d\n",$hour,$min,$sec;
	$log .= sprintf "loop end at %02d:%02d:%02d\n",$hour,$min,$sec;
	push(@feature,"intron_donor_accepter");
}
undef @gff_data;
#READING snps.txt
open(IN, "$snpsfile") or die "Can't open file:$snpsfile\n";
while(<IN>){
	$snps_raw .= $_;
}
close(IN);
$snps_raw =~ s/\x0D\x0A|\x0D|\x0A/\n/g;
@snps_data = (split(/\n/,$snps_raw));
@snps_data = grep(/^[^\#]/,@snps_data);#remove comment line
undef $snps_raw;
printf STDERR "snps num:%d\n",$#snps_data+1;
$log .= sprintf "snps num:%d\n",$#snps_data+1;

#SNPs list to hash

foreach (@snps_data){
	my $p = (split(/\t/,$_))[$position];
	if(!(defined $p)){die "ERROR option:-p $position is invalid value\n"}
	$snps_hash{$p} = $_;
}
#MATCHING
foreach $feat (@feature){
	${$feat.'_key'} = &matching($feat,\@{$feat.'data'},\%snps_hash);
}

#MAKING output directory
($day,$month) = (localtime(time))[3,4];
$month++;
$prefix_dir =sprintf("%s%02d%02d",$prefix_dir,$month,$day);
$outdir = $prefix_dir;
$num = 2;
while(-d $outdir){
	$outdir ="$prefix_dir-$num";
	$num++;
}

mkdir($outdir) or die "Can\'t make output directory:$outdir\n";
print STDERR "Output directory:$outdir\n";
$log .= "Output directory:$outdir\n";
#OUTPUT results of matching
foreach $feat(@feature){
	&output($feat,${$feat.'_key'});
}
#OUTPUT intron data
if($source){
	local $" = "\n";
	open (OUT,">$outdir/"."intron_donor_accepter_data.txt") or die "ERROR:Can\'t open data file about intron donar_accpter";
	print OUT "#Intron donor/accpter data from $source data in input GFF file\n";
	print OUT "@intron_donor_accepterdata\n";
	close OUT;
}
#
foreach my $posi (sort {$a <=> $b} keys %snps_hash){
	my $out = "$posi";
	foreach my $feat (@feature){
		$out .= "\t${snps_in_.$feat}{$posi}"
	}
	if(!(defined $snp_feat{$posi})){$snp_feat{$posi}="none"}
	if($snp_feat{$posi} eq "none"){
		$out .= "\t$snp_feat{$posi}";
	}else{
		$out .= "\t$snp_feat{$posi}\t$snps_hash{$posi}";
	}



	push(@stats,$out);
}
	local $" = "\t";
	open (OUT,">$outdir/"."stats.txt") or die "ERROR:Can\'t open data file about stats";
	print OUT "#position\t@feature\tcomment\n";
	local $" = "\n";
	print OUT "@stats\n";
	close OUT;

#Output log
open (LOG,">$outdir/"."log.txt") or die "ERROR:Can\'t open log file";
print LOG $log;
close LOG;



print STDERR "Program is finshed.\n";

######################################################
#Sub routin
######################################################

sub matching{
	#Recive SNPs and Annotation
	my($key,$data,$snps) = @_;
	
	($sec,$min,$hour) = (localtime(time))[0,1,2];
	printf STDERR "$key matching loop start at %02d:%02d:%02d\n",$hour,$min,$sec;
	$log .= sprintf "$key matching loop start at %02d:%02d:%02d\n",$hour,$min,$sec;

	my @start;
	my @end;
	my @name;
	#EXTRACT start, end and name of feature
	for ($i = 0; $i <= $#{$data}; $i++){
		($start[$i],$end[$i],$name[$i]) = (split(/\t/,${$data}[$i]))[3,4,8];
#		$name[$i] =~ /(Parent=[^,;]+)/;
#		$name[$i] = $1;
#		$name[$i] =~ s/Parent=//g;
	}

	printf STDERR "num of start, end and name of %s;%d,%d,%d \n",$key,$#start+1,$#end+1,$#name+1;
	$log .= sprintf "num of start, end and name of %s;%d,%d,%d \n",$key,$#start+1,$#end+1,$#name+1;
	my @match;
	my $percent = 0;
	my $lcount = 0;
	
	foreach my $posi (sort {$a <=> $b} keys %{$snps}){
		#matching loop one snp against list of one feature
		for($k = 0;$k <= $#start; $k++){
			if($start[$k] <= $posi && $posi <= $end[$k]){
				push(@match,"${$snps}{$posi}\t$name[$k]\t$start[$k]\t$end[$k]");
				${snps_in_.$key}{$posi}++;
				$snp_feat{$posi} .= "$name[$k] ";
			}
		}
		#Show progress bar
		if($lcount/$#snps_data*100 >= 0) {
			$bar =int($percent/2);
			print STDERR "\r";
			print STDERR "*"x$bar."[$percent%]";
			$percent = int($lcount/$#snps_data*100);
			#$t = $lcount/$#snps_data;
			#$mod = $lcount%($#snps_data/100);
			#print "$lcount\t$percent\t$mod\t$#snps_data\t$t\n";
		}
		$lcount++;
		if(!(defined ${snps_in_.$key}{$posi})){${snps_in_.$key}{$posi}=0}
		
		#if(!(defined $snp_feat{$posi})){$snp_feat{$posi}=0}
		#print "$posi\t$key\t${snps_in_.$key}{$posi}\t$snp_feat{$posi}\n";
	}
	print STDERR "\r";
	print STDERR "*"x$bar."[100%]";
	printf STDERR "\nnum of SNPs within %s:%d\n", $key, $#match+1;
	$log .= sprintf "num of SNPs within %s:%d\n", $key, $#match+1;
	($sec,$min,$hour) = (localtime(time))[0,1,2];
	printf STDERR "$key matching loop end at %02d:%02d:%02d\n",$hour,$min,$sec;
	$log .= sprintf "$key matching loop end at %02d:%02d:%02d\n\n",$hour,$min,$sec;
	return \@match;

}

sub output{
local $" = "\n";
my($name,$data,) = @_;
open (OUT,">$outdir/"."$name"."_match.txt") or die "ERROR:Can\'t open result file about $name";
print OUT "@{$data}\n";
close OUT;
}
